import os
import typer
import yaml

from bupy import (
 __app_name__
)

home_dir = str(Path.home())
libvirt_image_dir = os.path.join(home_dir, '/.local/share/libvirt/images')

default_config = [__app_name__]


def intialize_app():
    """
    Initialize config directory/file and ensure that
    $HOME/.local/share/libvirt/images exists
    """
    config_dir = xdg.save_config_path(__app_name__)
    config_file = os.path.join(config_dir, 'config.yaml')

    if not

    if not os.path.exists(libvirt_image_dir):
        os.makedirs(libvirt_image_dir)


def _create_config_file(config_dir: str) -> str:
    try:
        with config_dir.open("w") as file:
            config_parser.write(file)
    except OSError:
        return DB_WRITE_ERROR
    return SUCCESS



from rich import print as rprint
import socket
import shlex
import subprocess
from typing import List


def find_port() -> int:
    """
    Find a random open port
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("", 0))
    sock.listen(1)
    port = sock.getsockname()[1]
    sock.close()
    return port


def launch(ram: int, qcow: str, ign: str, ports: List[int]) -> int:
    """
    Launch FCOS with QEMU
    """

    port_list = set([])
    port_list.add(22)
    if ports is not None:
        port_list.update(ports)

    hostfwd_list = []
    port_maps = []

    for port in port_list:
        host_port = find_port()

        if port == 22:
            ssh_port = host_port

        paired_ports = f"{host_port}:{port}"
        port_maps.append(paired_ports)

        hostfwd_port = f"hostfwd=tcp::{host_port}-:{port}"
        hostfwd_list.append(hostfwd_port)

    joined_hostfwd = ",".join(hostfwd_list)

    qemu_cmd = f"qemu-system-x86_64 -m {ram} -enable-kvm -cpu host -snapshot -daemonize \
        -drive if=virtio,file={qcow} -name bupy-vm-{ssh_port} \
        -fw_cfg name=opt/com.coreos/config,file={ign} \
        -nic user,model=virtio,{joined_hostfwd}"
    qemu_cmd_split = shlex.split(qemu_cmd)

    rprint("Launching QEMU VM...")
    process = subprocess.Popen(qemu_cmd_split, start_new_session=True)
    qemu_pid = process.pid

    rprint(f"The PID is: {qemu_pid}")
    rprint("")
    rprint(f"The SSH port is: {ssh_port}")
    rprint(
        f"ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no core@localhost -p {ssh_port}"
    )
    rprint("")
    rprint("The port mappings are:")
    rprint("")
    for pairs in port_maps:
        rprint(pairs)
    rprint("")
    return 0



from pathlib import Path
from subprocess import Popen, PIPE, STDOUT
import sys
import typer


# Todo: Capture stderr and do something about it
def butane_to_ignition(bu: bytes) -> bytes:
    """
    Create Ignition JSON from Butane YAML
    """
    butane = "butane"
    butane_subprocess = Popen(butane, stdin=PIPE, stdout=PIPE, stderr=STDOUT)
    ignition_data = butane_subprocess.communicate(input=bu)[0]

    return ignition_data


def butane_encode(bu: str) -> bytes:

    if sys.stdin.isatty():
        with open(f"{bu}", "rb") as opened_file:
            butane_encoded = opened_file.read()
    else:
        butane_encoded = bu.encode()
    return butane_encoded


def butane_write(bu: bytes, file_name: str, force: bool, exit: bool) -> None:

    if not Path(file_name).is_file() or force:
        with open(f"{file_name}", "w") as f:
            f.write(bu.decode("utf-8"))
        if exit:
            raise typer.Exit(code=0)
    else:
        overwrite_it = typer.confirm(f"Overwrite {file_name}?")
        if not overwrite_it:
            raise typer.Abort()
        with open(f"{file_name}", "w") as f:
            f.write(bu.decode("utf-8"))
        if exit:
            raise typer.Exit(code=0)



# Exception handling derived from
# https://github.com/cidrblock/td4a/blob/master/td4a/models/exception_handler.py
# MIT License
# Copyright (c) 2017 Bradley A. Thornton

from jinja2 import meta, Environment, Template, StrictUndefined, Undefined
import ruamel.yaml as yaml
from typer import Exit
import re
import sys
import traceback


class TemplateException(Exception):
    def __init__(self, value) -> None:
        self.value = value
        super().__init__(self)

    def __str__(self):
        return repr(self.value)


class TemplateExceptionHandler(object):
    """
    Handle YAML and Jinja2 exceptions
    """

    def __init__(self, function):
        self.function = function
        self.error = (
            self.template_file
        ) = self.tback = self.exc_type = self.exc_value = self.exc_traceback = None
        self.error_map = {
            "ruamel.yaml.parser.ParserError": self.parser_error,
            "ruamel.yaml.constructor.ConstructorError": self.constructor_error,
            "ruamel.yaml.constructor.DuplicateKeyError": self.duplicate_key_error,
            "ruamel.yaml.scanner.ScannerError": self.scanner_error,
            "jinja2.exceptions": self.jinja_error,
        }

    def __call__(self, *args):
        try:
            return self.function(*args)
        except Exception as error:
            self.error = error
            self.exc_type, self.exc_value, self.exc_traceback = sys.exc_info()
            self.tback = traceback.extract_tb(self.exc_traceback)

            error_module = getattr(error, "__module__", None)
            if error_module:
                full_error = "%s.%s" % (error.__module__, self.exc_type.__name__)
            else:
                full_error = self.exc_type.__name__
            handler = self.error_map.get(full_error, self.error_map.get(error_module, self.unhandled))
            self.template_file = args[0]
            message = handler()
            raise TemplateException(message)

    def error_response(self, message, line_number):
        if line_number:
            error_payload = f"Issue processing {self.template_file}:{line_number} with the error message of: {message}"
        else:
            error_payload = (
                f"Issue processing {self.template_file} with the error message of: {message}"
            )
        return error_payload

    def constructor_error(self):
        line_number = self.error.problem_mark.line + 1
        message = next(x for x in str(self.error).splitlines() if x.startswith("found"))
        return self.error_response(message=message, line_number=line_number)

    def duplicate_key_error(self):
        line_number = self.error.problem_mark.line + 1
        message = next(x for x in str(self.error).splitlines() if x.startswith("found")).split(
            "with"
        )[0]
        return self.error_response(message=message, line_number=line_number)

    def jinja_error(self):
        message = str(self.error).replace("'ruamel.yaml.comments.CommentedMap object'", "Object")
        line_numbers = [x for x in self.tback if re.search("^<.*>$", x[0])]
        if line_numbers:
            line_number = line_numbers[0][1]
        else:
            line_number = "unknown"
        return self.error_response(message=message, line_number=line_number)

    def parser_error(self):
        line_number = self.error.problem_mark.line + 1
        messages = [x for x in str(self.error).splitlines() if x.startswith("expected")]
        if messages:
            message = messages[0]
        else:
            message = str(self.error)
        return self.error_response(message=message, line_number=line_number)

    def scanner_error(self):
        line_number = self.error.problem_mark.line + 1
        message = str(self.error).splitlines()[0]
        return self.error_response(message=message, line_number=line_number)

    def unhandled(self):
        print(self.exc_type, self.exc_value, self.exc_traceback, self.tback, self.error)
        line_numbers = [x for x in self.tback if re.search("^<.*>$", x[0])]
        if line_numbers:
            line_number = line_numbers[0][1]
        else:
            line_number = None
        message = self.error
        return self.error_response(message=message, line_number=line_number)


@TemplateExceptionHandler
def read_template_vars(file_path: str) -> dict:
    """
    Reads in template vars from a file
    """
    try:
        with open(file_path, "rb") as f:
            template_vars = yaml.load(f, yaml.SafeLoader)
        return template_vars
    except TemplateException as err:
        print(err)


@TemplateExceptionHandler
def verify_template(file_path: str) -> bool:
    """
    Verifies if the loaded file is a valid Jinja2 template
    """
    env = Environment()
    try:
        with open(file_path, "r") as f:
            env.parse(f.read())
        return True
    except TemplateException as err:
        print(err)
        return False


@TemplateExceptionHandler
def rendered_template(file_path: str, template_vars: dict) -> bytes:
    """
    Renders a Jinja2 template from a file
    """
    try:
        with open(file_path, "r") as f:
            rendered_template = Template(f.read()).render(template_vars)
        return rendered_template.encode("utf-8")
    except TemplateException as err:
        print(err)



from bupy import __app_name__, __version__
from rich import print as rprint
import os
import shutil
import typer
import xdg


class SilenceTemplateException:
    def __init__(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, ex_traceback):
        from bupy.template import TemplateException

        if ex_type is TemplateException:
            rprint(f"[bold red]Error:[/bold red] [red]{ex_value}[/red]")
            raise typer.Abort()


def find_binary(name: str):
    """
    Finds a binary on the system by name and checks if it is executable.
    """
    binary = shutil.which(name)

    return binary


def create_xdg_config_dir(name: str) -> str:
    """
    Creates the bupy config directory in $XDG_CONFIG_HOME/bupy
    """

    directory = xdg.BaseDirectory.save_config_path(name)
    if not os.path.exists(directory):
        xdg_config_dir = os.makedirs(directory)

    return xdg_config_dir


def version_callback(value: bool) -> None:
    if value:
        rprint(f"{__app_name__}: {__version__}")
        raise typer.Exit()



from collections import deque
from pathlib import Path
from rich import print as rprint
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.progress import wrap_file
import copy
import json
import lzma
import os
import shutil
import tempfile
import urllib3


class Token:
    def __init__(self, name, is_var=False) -> None:
        self.name = name
        self.is_var = is_var

    def get(self, data, args, **kwargs):

        if not self.is_var:
            return data.get(self.name, {})
        v = kwargs.get(self.name, None)
        if not v:
            v = args.popleft()
        return data.get(v, {})


class Fcos:

    lookups = {
        "disk": [
            Token("architectures"),
            Token("arch", True),
            Token("artifacts"),
            Token("platform", True),
            Token("formats"),
            Token("disk_format", True),
            Token("disk"),
        ],
        "disk_location": [
            Token(*x)
            for x in [
                ("architectures",),
                ("arch", True),
                ("artifacts",),
                ("platform", True),
                ("formats",),
                ("disk_format", True),
                ("disk",),
                ("location",),
            ]
        ],
        "disk_sha256": [
            Token(*x)
            for x in [
                ("architectures",),
                ("arch", True),
                ("artifacts",),
                ("platform", True),
                ("formats",),
                ("disk_format", True),
                ("disk",),
                ("sha256",),
            ]
        ],
    }

    def __init__(self, data):
        self._data = data

    def get_item(self, item_name, *args, **kwargs):
        cur_data = copy.deepcopy(self._data)
        deque(args)

        for item in self.lookups[item_name]:
            cur_data = item.get(cur_data, args, **kwargs)
        return cur_data

    def get_disk(self, *args, **kwargs):
        return self.get_item("disk_location", *args, **kwargs)

    def get_sha256(self, *args, **kwargs):
        return self.get_item("disk_sha256", *args, **kwargs)


def get_stream_data(
    fcos_stream="stable",
    fcos_json_url="https://builds.coreos.fedoraproject.org/streams/",
):
    """
    Grabs FCOS stream data off a URL
    """
    http = urllib3.PoolManager()
    resp = http.request("GET", fcos_json_url + fcos_stream + ".json")
    fcos_stream_data = json.loads(resp.data)

    return fcos_stream_data


def libvirt_image_path(libvirt_image_path=None):

    if libvirt_image_path is not None:
        local_libvirt_image_path = libvirt_image_path
    else:
        home = str(Path.home())
        libvirt_image_path = "/.local/share/libvirt/images"
        local_libvirt_image_path = home + libvirt_image_path
    return local_libvirt_image_path


# TODO: Add list disks support
# def list_disks():
#     fcos_disks =
#     for path in os.scandir(dir_path):
#         if path.is_file():
#             rprint(path.name)

# TODO: If you ctrl+c in the middle of the extract it leaves a incomplete VM qcow2 file. We should clean that up if we exit in the middle of an extraction
def download_disk(download_url, download_path=None):
    """
    Download and verify FCOS disk image
    """

    if download_path is not None:
        download_path = download_path
    else:
        download_path = libvirt_image_path()

    archive_name = os.path.basename(download_url)
    disk_name = os.path.splitext(archive_name)[0]
    disk_path = download_path + "/" + disk_name

    if not os.path.exists(download_path):
        os.makedirs(download_path, mode=0o775)
        rprint(download_path + " does not exist. Creating it!")

    if os.path.exists(disk_path):
        rprint(disk_path + " already exists. Skipping download!")
        rprint("")
    else:
        with tempfile.TemporaryDirectory() as tmpdirname:

            archive_path = tmpdirname + "/" + archive_name
            http = urllib3.PoolManager()
            download_file = http.request("GET", download_url, preload_content=False)
            download_size = int(download_file.headers["Content-Length"])

            with wrap_file(
                download_file,
                download_size,
                description=f"Downloading {archive_name}",
            ) as req:
                with open(archive_path, "wb") as out_file:
                    shutil.copyfileobj(req, out_file)

            with Progress(
                TextColumn("[progress.description]{task.description}"),
                SpinnerColumn(),
                transient=True,
            ) as progress:
                progress.add_task(description=f"Extracting {archive_name}...", total=None)
                with lzma.open(archive_path) as archive_content:
                    with open(disk_path, "wb") as write_disk:
                        shutil.copyfileobj(archive_content, write_disk)
            rprint(f"Extracting {archive_name}... Done!")

    return disk_path



from bupy import butane, fcos, qemu, util, __app_name__, __version__
from bupy import template as tpl
from pathlib import Path
from rich import print as rprint
from rich import print_json
from rich.syntax import Syntax
from rich.console import Console
from tempfile import NamedTemporaryFile
from typing import List, Optional
import sys
import time
import typer


app = typer.Typer(
    help="[bold]Bupy:[/bold] [bold]Bu[/bold]tane [bold]Py[/bold]thon toolkit.",
    rich_markup_mode="rich",
    no_args_is_help=True,
    epilog="Made in [pale_turquoise1]✶✶✶✶[/pale_turquoise1] [red1]Chicago[/red1][pale_turquoise1]✶✶✶✶[/pale_turquoise1]  〜 (c) 2022 QuickVM, LLC",
    pretty_exceptions_show_locals=False,
)


@app.callback()
def main(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        "-v",
        help="Show the version and exit.",
        callback=util.version_callback,
        is_eager=True,
    )
) -> None:
    return


@app.command()
def convert(
    file: str = typer.Argument(
        ... if sys.stdin.isatty() else sys.stdin.read().strip(),
        help="Reads Butane from a file or stdin and converts it to Ignition JSON.",
    ),
    write: str = typer.Option("", "--write", "-w", help="Write the Ignition JSON to a file."),
    force: bool = typer.Option(False, "--force", "-f", help="Overwrite an existing Ignition file."),
    pretty: bool = typer.Option(
        False, "--pretty", "-p", help="Makes the Ignition JSON pretty. Sooo pretty..."
    ),
):
    """
    Converts Butane YAML to Ignition JSON
    """
    butane_binary = util.find_binary("butane")

    if butane_binary is None:
        rprint(
            "[bold red]Error:[/bold red] The butane binary could not be found! Please [link=https://coreos.github.io/butane/getting-started/]install[/link] it."
        )
        raise typer.Exit(code=1)

    butane_file = butane.butane_encode(file)

    ignition_json = butane.butane_to_ignition(butane_file)

    if write:
        butane.butane_write(ignition_json, file, force, True)
    if pretty:
        print_json(ignition_json.decode("utf-8"))
    else:
        if sys.stdout.isatty():
            rprint(ignition_json.decode("utf-8"))
        else:
            print(ignition_json.decode("utf-8"))


# @app.command()
# def merge():
#     """
#     Merges Butane files together
#     """
#     typer.echo("Merging Butane files...")


@app.command()
def vm(
    file: str = typer.Argument(
        ... if sys.stdin.isatty() else sys.stdin.read().strip(),
        help="Reads Butane from a file or stdin and converts it to Ignition JSON.",
        show_default=False,
    ),
    template: str = typer.Option(
        "", "--template", "-t", help="Read file as a Jinja2 template and load a variables file."
    ),
    write: str = typer.Option("", "--write", "-w", help="Write the Ignition JSON to a file."),
    force: bool = typer.Option(False, "--force", "-f", help="Overwrite an existing Ignition file."),
    stream: str = typer.Option(
        "stable",
        "--stream",
        "-s",
        help="Fedora CoreOS Stream: stable, testing, next",
    ),
    arch: str = typer.Option(
        "x86_64",
        "--arch",
        "-a",
        help="Fedora CoreOS arch: aarch64, s390x, x86_64",
    ),
    ram: int = typer.Option(1024, "--ram", "-r", help="RAM amount in MB"),
    ports: Optional[List[int]] = typer.Option(
        None,
        "--port",
        "-p",
        help="Map random host ports to VM ports. Can be used many times.",
    ),
):
    """
    Launches a QEMU VM with a Butane YAML or Jinja2 Template
    """

    butane_binary = util.find_binary("butane")
    qemu_binary = util.find_binary("qemu-system-x86_64")

    if butane_binary is None:
        rprint(
            "[bold red]Error:[/bold red] The butane binary could not be found! Please [link=https://coreos.github.io/butane/getting-started/]install[/link] it."
        )
        raise typer.Exit(code=1)

    if qemu_binary is None:
        rprint(
            "[bold red]Error:[/bold red] The qemu-kvm binary could not be found! Please [link=https://www.qemu.org/download/]install[/link] it."
        )
        raise typer.Exit(code=1)

    # Download FCOS image and stick it in ~/.local/share/libvirt/images
    # TODO: move this to fcos.py and add download flags
    home = str(Path.home())
    libvirt_image_path = "/.local/share/libvirt/images"
    local_fcos_download_path = home + libvirt_image_path
    fcos_stream_data = fcos.get_stream_data(stream)
    fcos_download_url = fcos.Fcos(fcos_stream_data).get_disk(
        arch=arch, platform="qemu", disk_format="qcow2.xz"
    )
    fcos_disk = fcos.download_disk(fcos_download_url, local_fcos_download_path)

    if template:
        with util.SilenceTemplateException():
            butane_file = tpl.rendered_template(file, tpl.read_template_vars(template))
    else:
        butane_file = butane.butane_encode(file)

    ignition_json = butane.butane_to_ignition(butane_file)

    if write:
        butane.butane_write(ignition_json, file, force, False)

    ignition_launch_file = NamedTemporaryFile(delete=False, suffix=".ign")
    ignition_launch_file_path = ignition_launch_file.name

    with open(ignition_launch_file_path, "wb") as ignition_launch_file:
        ignition_launch_file.write(ignition_json)

    qemu.launch(ram, fcos_disk, ignition_launch_file_path, ports)
    rprint("VM Launched! :rocket:")
    time.sleep(2)
    Path(ignition_launch_file_path).unlink()


# @app.command()
# def serve():
#     """
#     Serve an Ignition file via HTTP on a specified port
#     """
#     typer.echo("Coming soon...")


@app.command()
def template(
    template: str = typer.Argument(
        ... if sys.stdin.isatty() else sys.stdin.read().strip(),
        help="Reads a Jinja2 template from a file or stdin, renders the template and converts it to Ignition JSON.",
        show_default=False,
    ),
    variables: str = typer.Argument(
        ...,
        help="Reads a YAML file to use as variables for rendering a Jinja2 template.",
    ),
    show: bool = typer.Option(
        False, "--show", "-s", help="rPrint the rendered template.", show_default=False
    ),
    line_numbers: bool = typer.Option(
        True, "--numbers", "-n", help="Show line numbers in printed template."
    ),
    write: str = typer.Option("", "--write", "-w", help="Write the Ignition JSON to a file."),
    force: bool = typer.Option(False, "--force", "-f", help="Overwrite an existing Ignition file."),
    pretty: bool = typer.Option(
        False, "--pretty", "-p", help="Makes the Ignition JSON pretty. Very pretty..."
    ),
):
    """
    Renders a Jinja2 Template to Butane YAML or Ignition JSON
    """
    butane_binary = util.find_binary("butane")

    if butane_binary is None:
        rprint(
            "[bold red]Error:[/bold red] The butane binary could not be found! Please [link=https://coreos.github.io/butane/getting-started/]install[/link] it."
        )
        raise typer.Exit(code=1)

    with util.SilenceTemplateException():
        if show:
            bu = tpl.rendered_template(template, tpl.read_template_vars(variables)).decode("utf-8")
            syntax = Syntax(bu, "YAML", line_numbers=line_numbers)
            console = Console(color_system=None)
            console.print(syntax)
            raise typer.Exit()

        # TODO: Add support for finding bupyvars.yaml file in CWD, or in the directory the template resides in.
        # cwd = Path.cwd()
        template_output = tpl.rendered_template(template, tpl.read_template_vars(variables))
        ignition_json = butane.butane_to_ignition(template_output)

    if write:
        butane.butane_write(ignition_json, write, force, True)
    if pretty:
        print_json(ignition_json.decode("utf-8"))
    else:
        if sys.stdout.isatty():
            rprint(ignition_json.decode("utf-8"))
        else:
            print(ignition_json.decode("utf-8"))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bupy']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'pyxdg>=0.28,<0.29',
 'rich>=12.5.1,<13.0.0',
 'ruamel.yaml>=0.17.21,<0.18.0',
 'typer[all]>=0.6.1,<0.7.0',
 'urllib3>=1.26.10,<2.0.0']

entry_points = \
{'console_scripts': ['bupy = bupy.cli:app']}

setup_kwargs = {
    'name': 'bupy',
    'version': '0.1.0',
    'description': 'A Python toolkit for Butane and Ignition',
    'long_description': '# Bupy\n\nThe **Bu**tane **Py**thon Toolkit\n\nBupy was built to help users develop Butane configurations for Fedora CoreOS quickly on their workstations. It allows you to convert Butane YAML to Ignition JSON and render Butane Jinja2 templates to either Butane YAML or Ignition JSON. You can then use your Butane file or template to launch a local QEMU Virtual Machine.\n\n### Roadmap\n[x] Convert Support\n[x] Jinja2 Template Support\n[x] Launch a local QEMU FCOS VM\n[] Merge Butane YAML (snippets)\n[] Serve Ignition JSON via HTTP\n[] Libvirt support\n\n### Demo\n\nYou can watch a quick demo of Bupy on Youtube.\n\n[![Quick demo of Bupy for Fedora CoreOS](https://img.youtube.com/vi/yBOEz827TUU/0.jpg)](https://www.youtube.com/watch?v=yBOEz827TUU)\n\n### Development\n\n1) Clone this repo\n1) Install dependencies\n  ```\n  poetry install\n  ```\n1) Activate a poetry shell\n  ```\n  $ poetry shell\n  Spawning shell within /home/jdoss/src/quickvm/bupy/.venv\n  . /home/jdoss/src/quickvm/bupy/.venv/bin/activate\n  $ . /home/jdoss/src/quickvm/bupy/.venv/bin/activate\n  ```\n1) Make changes...\n1) See them in action\n  ```\n  (.venv) $ python -m bupy --help\n\n   Usage: bupy [OPTIONS] COMMAND [ARGS]...\n\n  ╭─ Options ────────────────────────────────────────────────────────────────────────────────────────╮\n  │ --version             -v        Show the version and exit.                                       │\n  │ --install-completion            Install completion for the current shell.                        │\n  │ --show-completion               Show completion for the current shell, to copy it or customize   │\n  │                                 the installation.                                                │\n  │ --help                          Show this message and exit.                                      │\n  ╰──────────────────────────────────────────────────────────────────────────────────────────────────╯\n  ╭─ Commands ───────────────────────────────────────────────────────────────────────────────────────╮\n  │ convert      Converts Butane YAML to Ignition JSON                                               │\n  │ merge        Merge Butane files together                                                         │\n  │ qemu         Launches a QEMU VM with the specified Ignition JSON or Butane YAML                  │\n  │ serve        Serve an ignition file via HTTP on a specified port                                 │\n  │ template     Render Butane Jinja2 templates                                                      │\n  ╰──────────────────────────────────────────────────────────────────────────────────────────────────╯\n  ```\n',
    'author': 'QuickVM',
    'author_email': 'hello@quickvm.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
